// live.cpp : Defines the entry point for the application.

#include "stdafx.h"
#include "live.h"
#include <string>
#include <EGrabber.h>
#include <RGBConverter.h>

using namespace Euresys;

typedef EGrabber<CallbackOnDemand> MyGrabber;

static const size_t     NB_BUFFERS     = 50;
static const size_t     MAX_LOADSTRING = 100;

// Global Variables:
static HINSTANCE        hInst;                          // Current instance
static TCHAR            szTitle[MAX_LOADSTRING];        // The title bar text
static TCHAR            szWindowClass[MAX_LOADSTRING];  // The main window class name

// Forward declarations of functions included in this code module:
struct LiveState;
static ATOM             MyRegisterClass(HINSTANCE hInstance);
static VOID             InitInstance(HINSTANCE, int, LiveState *);
static LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
static void             OnPaint(HWND hWnd, LiveState *);
static DWORD WINAPI     RefreshThread(LPVOID context);

// Application State
struct LiveState {
    Euresys::GenTL          genTL;
    MyGrabber               grabber;
    RGBConverter            rgbConverter;
    BITMAPINFO              *bitmapInfo;
    Euresys::Buffer         *currentBuffer;
    HWND                    window;
    UINT_PTR                refreshTimer;
    HANDLE                  refreshThread;
    volatile bool           refreshThreadRunning;
    size_t                  width;
    size_t                  height;
    bool                    imageConvertFailed;

    LiveState()
    : grabber(genTL), rgbConverter(genTL), bitmapInfo(NULL), currentBuffer(NULL)
    , window(NULL), refreshTimer(0), refreshThread(NULL)
    , refreshThreadRunning(false), width(0), height(0)
    , imageConvertFailed(false)
    {
        grabber.runScript("./config.js");
        grabber.reallocBuffers(NB_BUFFERS);
        width = grabber.getWidth();
        height = grabber.getHeight();
        bitmapInfo = (LPBITMAPINFO)malloc(sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD));
        bitmapInfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
        bitmapInfo->bmiHeader.biPlanes = 1;
        bitmapInfo->bmiHeader.biBitCount = 24;
        bitmapInfo->bmiHeader.biCompression = BI_RGB;
        bitmapInfo->bmiHeader.biSizeImage = 0;
        bitmapInfo->bmiHeader.biXPelsPerMeter = 0;
        bitmapInfo->bmiHeader.biYPelsPerMeter = 0;
        bitmapInfo->bmiHeader.biClrUsed = 0;
        bitmapInfo->bmiHeader.biClrImportant = 0;
        bitmapInfo->bmiHeader.biWidth = (LONG)width;
        bitmapInfo->bmiHeader.biHeight = -(LONG)height;
        RGBQUAD* bmiColors = (RGBQUAD*)(bitmapInfo->bmiColors);
        for (size_t index = 0; index < 1; ++index)
        {
            bmiColors[index].rgbBlue = (BYTE)index;
            bmiColors[index].rgbGreen = (BYTE)index;
            bmiColors[index].rgbRed = (BYTE)index;
            bmiColors[index].rgbReserved = 0;
        }
    }

    virtual ~LiveState()
    {
        stopRefreshThread();
        if (currentBuffer)
        {
            currentBuffer->push(grabber);
            delete currentBuffer;
        }
        free(bitmapInfo);
    }

    void startRefreshThread()
    {
        refreshThreadRunning = true;
        refreshThread = CreateThread(NULL, 0, RefreshThread, this, 0, NULL);
        refreshTimer = SetTimer(window, 0, 50, (TIMERPROC) NULL);
    }

    void stopRefreshThread()
    {
        if (refreshTimer)
        {
            KillTimer(window, refreshTimer);
            refreshTimer = 0;
        }
        refreshThreadRunning = false;
        if (refreshThread)
        {
            try
            {
                grabber.stop();
                grabber.cancelPop();
            }
            catch (...)
            {
            }
            WaitForSingleObject(refreshThread, INFINITE);
            CloseHandle(refreshThread);
            refreshThread = NULL;
        }
    }
};

int APIENTRY _tWinMain(HINSTANCE hInstance,
                       HINSTANCE hPrevInstance,
                       LPTSTR    lpCmdLine,
                       int       nCmdShow)
{
    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);

    try
    {
        LiveState liveState;

        // TODO: Place code here.
        MSG msg;
        HACCEL hAccelTable;

        // Initialize global strings
        LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
        LoadString(hInstance, IDC_LIVE, szWindowClass, MAX_LOADSTRING);
        MyRegisterClass(hInstance);

        // Perform application initialization:
        InitInstance(hInstance, nCmdShow, &liveState);
        hAccelTable = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDC_LIVE));

        // Main message loop:
        BOOL bRet;
        while ((bRet = GetMessage(&msg, NULL, 0, 0)) != 0)
        {
            if (bRet == -1)
            {
                throw std::runtime_error("Handle error in GetMessage");
            }
            if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
        return (int) msg.wParam;
    }
    catch (const std::exception &e)
    {
        MessageBoxA(NULL, e.what(), NULL, MB_OK);
        return FALSE;
    }
}



//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
    WNDCLASSEX wcex;

    wcex.cbSize         = sizeof(WNDCLASSEX);
    wcex.style          = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc    = WndProc;
    wcex.cbClsExtra     = 0;
    wcex.cbWndExtra     = 0;
    wcex.hInstance      = hInstance;
    wcex.hIcon          = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_LIVE));
    wcex.hCursor        = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground  = (HBRUSH)(COLOR_WINDOW+1);
    wcex.lpszMenuName   = MAKEINTRESOURCE(IDC_LIVE);
    wcex.lpszClassName  = szWindowClass;
    wcex.hIconSm        = LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_SMALL));

    return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HINSTANCE, int, LiveState *)
//
//   PURPOSE: Creates main window
//
//   COMMENTS:
//
//        In this function, we create and display the main program window.
//
VOID InitInstance(HINSTANCE hInstance, int nCmdShow, LiveState *liveState)
{
    std::wstringstream sstitle;
    sstitle << szTitle << " " << liveState->width << "x" << liveState->height;

    liveState->window = CreateWindow(szWindowClass, sstitle.str().c_str(), WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, 0, (int)liveState->width, (int)liveState->height, NULL, NULL, hInstance, liveState);

    if (!liveState->window)
    {
        throw std::runtime_error("CreateWindow failed");
    }

    liveState->grabber.start();
    liveState->startRefreshThread();

    ShowWindow(liveState->window, nCmdShow);
    UpdateWindow(liveState->window);
}

//
//  FUNCTION: WndProc(HWND, UINT, WPARAM, LPARAM)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND    - process the application menu
//  WM_PAINT    - Paint the main window
//  WM_DESTROY    - post a quit message and return
//
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    LiveState *pliveState;
    if (message == WM_NCCREATE)
    {
        CREATESTRUCT *pCreate = reinterpret_cast<CREATESTRUCT*>(lParam);
        pliveState = reinterpret_cast<LiveState*>(pCreate->lpCreateParams);
        SetWindowLongPtr(hWnd, GWLP_USERDATA, (LONG_PTR)pliveState); 
        return DefWindowProc(hWnd, message, wParam, lParam);
    }
    else
    {
        pliveState = reinterpret_cast<LiveState*>(GetWindowLongPtr(hWnd, GWLP_USERDATA));
        if (pliveState == NULL)
        {
            return DefWindowProc(hWnd, message, wParam, lParam);
        }
    }
    LiveState &liveState = *pliveState;

    switch (message)
    {
    case WM_COMMAND:
        {
            int wmId    = LOWORD(wParam);
            int wmEvent = HIWORD(wParam);
            // Parse the menu selections:
            switch (wmId)
            {
            case IDM_EXIT:
                DestroyWindow(hWnd);
                break;
            default:
                return DefWindowProc(hWnd, message, wParam, lParam);
            }
        }
        break;
    case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hWnd, &ps);
            OnPaint(hWnd, pliveState);
            EndPaint(hWnd, &ps);
        }
        break;
    case WM_GETMINMAXINFO:
        if (0 == DefWindowProc(hWnd, message, wParam, lParam))
        {
            MINMAXINFO* pmmi = (MINMAXINFO*)lParam;
            pmmi->ptMaxTrackSize.x = (LONG)liveState.width;
            pmmi->ptMaxTrackSize.y = (LONG)liveState.height;
            pmmi->ptMinTrackSize.x = (LONG)liveState.width;
            pmmi->ptMinTrackSize.y = (LONG)liveState.height;
        }
        break;
    case WM_TIMER:
        PostMessage(hWnd, WM_PAINT, 0, 0);
        break;
    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
    }
    return 0;
}

size_t GetBitmapRowSize(size_t width, int bpp)
{
    // row size must be a multiple of 4 bytes
    size_t rowSize = width * bpp;
    if (rowSize & 3) {
        rowSize |= 3;
        ++rowSize;
    }
    return rowSize;
}

void OnPaint(HWND hWnd, LiveState *pliveState)
{
    LiveState &liveState = *pliveState;
    if (!liveState.currentBuffer || liveState.imageConvertFailed)
    {
        return;
    }
    unsigned char *imagePtr = liveState.currentBuffer->getInfo<unsigned char *>(liveState.grabber, gc::BUFFER_INFO_BASE);
    const std::string format(liveState.genTL.imageGetPixelFormat(liveState.currentBuffer->getInfo<uint64_t>(liveState.grabber, gc::BUFFER_INFO_PIXELFORMAT)));
    const size_t width = liveState.currentBuffer->getInfo<size_t>(liveState.grabber, gc::BUFFER_INFO_WIDTH);
    const size_t height = liveState.currentBuffer->getInfo<size_t>(liveState.grabber, gc::BUFFER_INFO_DELIVERED_IMAGEHEIGHT);
    try
    {
        RGBConverter::BGR8 bgr(liveState.rgbConverter, imagePtr, format, width, height);
        unsigned char *imgPtr = bgr.getBuffer();
        size_t lineDataSize = liveState.width * 3;
        size_t rowSize = GetBitmapRowSize(liveState.width, 3);
        HDC dc = GetDC(hWnd);
        if (rowSize /= lineDataSize)
        {
            for (size_t h = 0; h < liveState.height; ++h) {
                ::SetDIBitsToDevice(dc, 0, (int)h, (DWORD)liveState.width, 1, 0, 0, 0, 1, imgPtr, liveState.bitmapInfo, DIB_RGB_COLORS);
                imgPtr += lineDataSize;
            }
        }
        else
        {
            ::SetDIBitsToDevice(dc, 0, 0, (DWORD)liveState.width, (DWORD)liveState.height, 0, 0, 0, (UINT)liveState.height, imgPtr, liveState.bitmapInfo, DIB_RGB_COLORS);
        }
        ReleaseDC(hWnd, dc);
    }
    catch (const std::exception &e)
    {
        liveState.imageConvertFailed = true;
        MessageBoxA(NULL, e.what(), NULL, MB_ICONSTOP | MB_OK | MB_TASKMODAL);
        PostQuitMessage(0);
        return;
    }
    Euresys::Buffer *p = liveState.currentBuffer;
    liveState.currentBuffer = NULL;
    p->push(liveState.grabber);
    delete p;
}

DWORD WINAPI RefreshThread(LPVOID context)
{
    LiveState &liveState = *(LiveState*)context;
    while (liveState.refreshThreadRunning)
    {
        try
        {
            Euresys::Buffer buffer(liveState.grabber.pop());
            if (liveState.currentBuffer)
            {
                buffer.push(liveState.grabber);
            }
            else
            {
                liveState.currentBuffer = new Euresys::Buffer(buffer);
            }
        }
        catch (const Euresys::gentl_error &e)
        {
            if (e.gc_err != GenICam::Client::GC_ERR_ABORT)
            {
                MessageBoxA(NULL, e.what(), NULL, MB_OK);
            }
        }
        catch (const std::exception &e)
        {
            MessageBoxA(NULL, e.what(), NULL, MB_OK);
        }
    }
    return 1;
}
