﻿using System;

namespace egrabber
{
    class Program
    {
        const int BUFFER_COUNT = 5;
        const int FRAME_COUNT = 20;

        static void sampleCallbackOnDemand(Euresys.GenTL genTL)
        {
            using (Euresys.EGrabberCallbackOnDemand grabber = new Euresys.EGrabberCallbackOnDemand(genTL))
            {
                using (Euresys.RGBConverter.RGBConverter rgbConverter = new Euresys.RGBConverter.RGBConverter(genTL))
                {
                    System.Console.WriteLine();
                    System.Console.WriteLine("Using {0}", grabber.GetType().FullName);
                    System.Console.WriteLine();
                    // grabber.runScript("config.js");
                    grabber.reallocBuffers(BUFFER_COUNT);

                    String str;
                    grabber.getInfoDeviceModule(Euresys.gc.DEVICE_INFO_CMD.DEVICE_INFO_DISPLAYNAME, out str);
                    System.Console.WriteLine("DEVICE_INFO_DISPLAYNAME = {0}", str);
                    grabber.getInfoSystemModule(Euresys.gc.TL_INFO_CMD.TL_INFO_PATHNAME, out str);
                    System.Console.WriteLine("TL_INFO_PATHNAME = {0}", str);

                    String[] features = grabber.getStringListRemoteModule(Euresys.GenApiQueryBuilder.Features());
                    System.Console.WriteLine("Remote device features:");
                    bool isExposureModeAvailable = false;
                    foreach (String f in features)
                    {
                        System.Console.WriteLine(" - {0}", f);
                        if (f == "ExposureMode") {
                            isExposureModeAvailable = true;
                        }
                    }

                    if (isExposureModeAvailable)
                    {
                        String[] entries = grabber.getStringListRemoteModule(Euresys.GenApiQueryBuilder.EnumEntries("ExposureMode"));
                        System.Console.WriteLine("ExposureMode entries:");
                        foreach (String ee in entries)
                        {
                            System.Console.WriteLine(" - {0}", ee);
                        }
                    }

                    UInt64 width = grabber.getWidth();
                    UInt64 height = grabber.getHeight();
                    String format = grabber.getPixelFormat();
                    System.Console.WriteLine("Grabber config: {0}x{1}, pixel format: {2}", width, height, format);

                    grabber.start(FRAME_COUNT);
                    for (ulong frame = 0; frame < FRAME_COUNT; ++frame)
                    {
                        using (Euresys.ScopedBuffer buffer = new Euresys.ScopedBuffer(grabber, 1000))
                        {
                            Euresys.SizeT size;
                            UInt64 timestamp;
                            IntPtr imgPtr;
                            UInt64 intFormat;
                            buffer.getInfo(Euresys.gc.BUFFER_INFO_CMD.BUFFER_INFO_SIZE_FILLED, out size);
                            buffer.getInfo(Euresys.gc.BUFFER_INFO_CMD.BUFFER_INFO_TIMESTAMP, out timestamp);
                            buffer.getInfo(Euresys.gc.BUFFER_INFO_CMD.BUFFER_INFO_BASE, out imgPtr);
                            buffer.getInfo(Euresys.gc.BUFFER_INFO_CMD.BUFFER_INFO_PIXELFORMAT, out intFormat);
                            System.Console.WriteLine("Acquisition Index {0} size={1} timestamp={2}.{3} imagePointer={4}",
                                frame, size, timestamp / 1000000, timestamp % 1000000, imgPtr);

                            using (Euresys.RGBConverter.BGR8 bgr = new Euresys.RGBConverter.BGR8(rgbConverter, imgPtr, format, width, height))
                            {
                                System.Console.WriteLine("bgr data: {0}", bgr.getBuffer());
                            }
                        }
                    }
                }
            }
        }

        static void sampleMultiThread(Euresys.GenTL genTL)
        {
            using (Euresys.EGrabberCallbackMultiThread grabber = new Euresys.EGrabberCallbackMultiThread(genTL))
            {
                System.Console.WriteLine();
                System.Console.WriteLine("Using {0}", grabber.GetType().FullName);
                System.Console.WriteLine();
                try
                {
                    grabber.enableAllEvent();
                    grabber.runScript("var p = grabbers[0].StreamPort; for (var s of p.$ee('EventSelector')) {p.set('EventNotification['+s+']', true);}");
                    grabber.reallocBuffers(BUFFER_COUNT);
                    grabber.onNewBufferEvent = delegate(Euresys.EGrabberCallbackMultiThread g, Euresys.NewBufferData data)
                    {
                        using (Euresys.ScopedBuffer buffer = new Euresys.ScopedBuffer(g, data))
                        {
                            System.Console.WriteLine("{0} | NEW BUFFER", data.timestamp);
                        }
                    };
                    grabber.onDataStreamEvent = delegate(Euresys.EGrabberCallbackMultiThread g, Euresys.DataStreamData data)
                    {
                        System.Console.WriteLine("{0} | {1}", data.timestamp, data.numid.ToString());
                    };
                    grabber.start();
                    System.Console.WriteLine("grabber started, sleeping...");
                    System.Threading.Thread.Sleep(100);
                    System.Console.WriteLine("grabber cleanup...");
                }
                finally
                {
                    grabber.runScript("var p = grabbers[0].StreamPort; for (var s of p.$ee('EventSelector')) {p.set('EventNotification['+s+']', false);}");
                }
            }
        }

        static void sampleSingleThread(Euresys.GenTL genTL)
        {
            using (Euresys.EGrabberCallbackSingleThread grabber = new Euresys.EGrabberCallbackSingleThread(genTL))
            {
                System.Console.WriteLine();
                System.Console.WriteLine("Using {0}", grabber.GetType().FullName);
                System.Console.WriteLine();
                try
                {
                    grabber.enableAllEvent();
                    grabber.runScript("var p = grabbers[0].StreamPort; for (var s of p.$ee('EventSelector')) {p.set('EventNotification['+s+']', true);}");
                    grabber.reallocBuffers(BUFFER_COUNT);
                    grabber.onNewBufferEvent = delegate(Euresys.EGrabberCallbackSingleThread g, Euresys.NewBufferData data)
                    {
                        using (Euresys.ScopedBuffer buffer = new Euresys.ScopedBuffer(g, data))
                        {
                            System.Console.WriteLine("{0} | NEW BUFFER", data.timestamp);
                        }
                    };
                    grabber.onDataStreamEvent = delegate(Euresys.EGrabberCallbackSingleThread g, Euresys.DataStreamData data)
                    {
                        System.Console.WriteLine("{0} | {1}", data.timestamp, data.numid.ToString());
                    };
                    grabber.start();
                    System.Console.WriteLine("grabber started, sleeping...");
                    System.Threading.Thread.Sleep(100);
                    System.Console.WriteLine("grabber cleanup...");
                }
                finally
                {
                    grabber.runScript("var p = grabbers[0].StreamPort; for (var s of p.$ee('EventSelector')) {p.set('EventNotification['+s+']', false);}");
                }
            }
        }

        public class EGrabberCallbacksWorker
        {
            public EGrabberCallbacksWorker(Euresys.EGrabberCallbackOnDemand grabber)
            {
                this.grabber = grabber;
            }
            public void DoWork()
            {
                try
                {
                    while (!_shouldStop)
                    {
                        ulong n = grabber.processEventFilter(Euresys.EventSelector.Any);
                        System.Console.WriteLine("  pending event count: {0}", n);
                    }
                }
                catch (System.Exception e)
                {
                    System.Console.WriteLine("processEventFilter: Exception = {0}", e.Message);
                }
                finally
                {
                    System.Console.WriteLine("EGrabberCallbacksWorker completed");
                }
            }
            public void RequestStop()
            {
                _shouldStop = true;
                grabber.cancelEventFilter(Euresys.EventSelector.Any);
            }
            private volatile bool _shouldStop;
            Euresys.EGrabberCallbackOnDemand grabber;
        }
        static void sampleCallbackOnDemandFromThread(Euresys.GenTL genTL)
        {
            using (Euresys.EGrabberCallbackOnDemand grabber = new Euresys.EGrabberCallbackOnDemand(genTL))
            {
                System.Console.WriteLine();
                System.Console.WriteLine("Using {0}", grabber.GetType().FullName);
                System.Console.WriteLine();
                try
                {
                    grabber.enableAllEvent();
                    grabber.runScript("var p = grabbers[0].StreamPort; for (var s of p.$ee('EventSelector')) {p.set('EventNotification['+s+']', true);}");
                    grabber.reallocBuffers(BUFFER_COUNT);
                    grabber.onNewBufferEvent = delegate(Euresys.EGrabberCallbackOnDemand g, Euresys.NewBufferData data)
                    {
                        using (Euresys.ScopedBuffer buffer = new Euresys.ScopedBuffer(g, data))
                        {
                            System.Console.WriteLine("{0} | NEW BUFFER", data.timestamp);
                        }
                    };
                    grabber.onDataStreamEvent = delegate(Euresys.EGrabberCallbackOnDemand g, Euresys.DataStreamData data)
                    {
                        System.Console.WriteLine("{0} | {1}", data.timestamp, data.numid.ToString());
                    };
                    EGrabberCallbacksWorker worker = new EGrabberCallbacksWorker(grabber);
                    System.Threading.Thread workerThread = new System.Threading.Thread(worker.DoWork);
                    workerThread.Start();

                    grabber.start(2); // 2 frames
                    System.Console.WriteLine("grabber started, sleeping...");
                    System.Threading.Thread.Sleep(1000); // 1 second to make sure processEventFitler blocks
                    System.Console.WriteLine("grabber cleanup...");

                    worker.RequestStop();
                    workerThread.Join();
                }
                finally
                {
                    grabber.runScript("var p = grabbers[0].StreamPort; for (var s of p.$ee('EventSelector')) {p.set('EventNotification['+s+']', false);}");
                }

            }
        }

        static void Main(string[] args)
        {
            try
            {
                using (Euresys.GenTL genTL = new Euresys.GenTL())
                {
                    sampleCallbackOnDemand(genTL);
                    sampleMultiThread(genTL);
                    sampleSingleThread(genTL);
                    sampleCallbackOnDemandFromThread(genTL);
                }
            }
            catch (System.Exception e)
            {
                System.Console.WriteLine(e.Message);
            }
        }
    }
}
